function [p,Plast,a,q] = QRbanded(A,r)

% Implements inversion algorithm for two-sided banded matrices
% from Theorem 4.4 in manuscript "Computation of
% quasiseparable representations of Green matrices".
% The algorithm computes lower Green generators of the
% inverse matrix.
%
% Input: A banded matrix to be inverted,
%        r semi-bandwidth
%          (i.e., a tridiagonal matrix has r = 1)

% Check size of A
[Nr,Nc] = size(A);
if Nr == Nc
    N = Nr;
else
    disp('Matrix is not square')
    return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1: compute generators of UR decomposition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here R is upper triangular and upper banded
% with bandwidth 2r

R = zeros(N);
U = zeros(r+1,r+1,N-1);

% 1.1
Y = A(1:r,1:2*r+1);

% 1.2
% for k = 1:N-2*r-1
for k = 1:N-2*r
    Delta = [Y(:,1); A(k+r,k)];
    [Utemp,Rtemp] = qr(Delta);
    U(:,:,k) = Utemp; % this is (r+1)x(r+1)
    x = Rtemp(1);
    Z = Utemp'*[Y(:,2:2*r), zeros(r,1);...
        A(k+r,k+1:k+2*r-1), A(k+r,k+2*r)];  % (r+1)x2r
    X = Z(1,:);
    Y = Z(2:r+1,:);
    R(k,k) = x;
    R(k,k+1:k+2*r) = X;
end

% 1.3
% for k = N-2*r:N-r-1
for k = N-2*r+1:N-r-1
    Delta = [Y(:,1); A(k+r,k)];
    [Utemp,Rtemp] = qr(Delta);
    U(:,:,k) = Utemp;
    x = Rtemp(1);
    Z = Utemp'*[Y(:,2:N-k+1); A(k+r,k+1:N)];
    X = Z(1,:);
    Y = Z(2:r+1,:);
    R(k,k) = x;
    R(k,k+1:N) = X;
end
ZZ = Z;  % save for later (this is Z_{N-r-1} of size (r+1)x(r+1))
% last step
k = N-r;
Delta = [Y(:,1); A(k+r,k)];
[Utemp,Rtemp] = qr(Delta);
U(:,:,k) = Utemp;
x = Rtemp(1);
Z = Utemp'*[Y(:,2:N-k+1); A(k+r,k+1:N)];
X = Z(1,:);
Y = Z(2:r+1,:);
R(k,k) = x;
R(k,k+1:N) = X;

% 1.4
for k = N-r+1:N-1
    Delta = Y(:,1);
    [Utemp,Rtemp] = qr(Delta);
    U(1:N-k+1,1:N-k+1,k) = Utemp;
    x = Rtemp(1);
    Z = Utemp'*Y(:,2:N-k+1);
    X = Z(1,:);
    Y = Z(2:N-k+1,:);
    R(k,k) = x;
    R(k,k+1:N) = X;
end
R(N,N) = Y;

%%%%%% Check (optional) %%%%%
% B = R;
% for k = N-1:-1:N-r+1
%     sizeU = N-k+1;
%     B(k:N,:) = U(1:sizeU,1:sizeU,k)*B(k:N,:);
% end
% for k = N-r:-1:1
%     B(k:k+r,:) = U(1:r+1,1:r+1,k)*B(k:k+r,:);
% end
% disp(norm(A-B))
%%%%%% End check %%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2: compute lower Green generators of inv(A)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Extract lower Green generators of U*
pU = zeros(N-1,r);  % i-th row is pU(i)
dU = zeros(N-r,1);  % i-th entry is dU(i)
aU = zeros(r,r,N-1);% i-th matrix is aU(i)
qU = zeros(r,N-r);  % i-th column is qU(i)
% pU(N-r+1) is an rxr matrix and will be defined last
for k = 1:N-r
    Ukstar = U(:,:,k)';
    pU(k,:) = Ukstar(1,1:r);
    dU(k) = Ukstar(1,r+1);
    aU(:,:,k) = Ukstar(2:r+1,1:r);
    qU(:,k) = Ukstar(2:r+1,r+1);
end

for k = N-r+1:N-1
    sizeU = N-k+1;
    Ukstar = U(1:sizeU,1:sizeU,k)';
    pU(k,1:sizeU) = Ukstar(1,:);
    aU(1:sizeU-1,1:sizeU,k) = Ukstar(2:sizeU,:);
end


% 2.1
q = qU;
a = aU;
p = zeros(N-1,r);

Ptemp = 1/R(N,N);
for k = N-1:-1:N-r+1
    p(k,1:N-k+1) = (pU(k,1:N-k+1)...
        -R(k,k+1:N)*Ptemp*aU(1:N-k,1:N-k+1,k))/R(k,k);
    Ptemp = [p(k,1:N-k+1); Ptemp*aU(1:N-k,1:N-k+1,k)];
end
Plast = Ptemp;  % computed as in lower banded case

% 2.2.1

for k = N-r:-1:N-2*r+1
    p(k,:) = (pU(k,:)-R(k,k+1:N)*Ptemp*a(:,:,k))/R(k,k);
    Ptemp = [p(k,:);Ptemp*a(:,:,k)];
end

% 2.2.2
t = Ptemp;
for k = N-2*r:-1:1
    p(k,:) = (pU(k,:)-R(k,k+1:k+2*r)*t*a(:,:,k))/R(k,k);
    t = [p(k,:); t(1:2*r-1,:)*a(:,:,k)];
end

end
    
    


